"""Part offer class."""
# Standard Modules
from dataclasses import dataclass
from typing import Dict

# Local Modules
from cofactr.schema.flagship.offer import Offer as FlagshipOffer
from cofactr.schema.types import OfferCorrection


@dataclass
class Offer(FlagshipOffer):
    """Part offer."""

    corrections: Dict[str, OfferCorrection]



"""Schema definitions."""
# Standard Modules
from enum import Enum
from typing import Callable, Dict

# Local Modules
from cofactr.helpers import identity
from cofactr.schema.flagship import (
    FlagshipOffer,
    FlagshipOrderStatus,
    FlagshipPart,
    FlagshipSeller,
)
from cofactr.schema.flagship_v2 import FlagshipV2Offer, FlagshipV2Part, FlagshipV2Seller
from cofactr.schema.flagship_v3 import FlagshipV3Offer, FlagshipV3Part, FlagshipV3Seller
from cofactr.schema.flagship_v4 import FlagshipV4Part, FlagshipV4Offer, FlagshipV4Seller
from cofactr.schema.flagship_v5 import FlagshipV5Offer, FlagshipV5Part, FlagshipV5Seller
from cofactr.schema.flagship_v6 import FlagshipV6Offer, FlagshipV6Part, FlagshipV6Seller
from cofactr.schema.flagship_v7 import FlagshipV7Offer, FlagshipV7Part
from cofactr.schema.flagship_alts_v0 import FlagshipAltsV0Part
from cofactr.schema.flagship_cache_v0 import FlagshipCacheV0Part
from cofactr.schema.flagship_cache_v1 import FlagshipCacheV1Part
from cofactr.schema.flagship_cache_v2 import FlagshipCacheV2Part
from cofactr.schema.flagship_cache_v3 import FlagshipCacheV3Part
from cofactr.schema.flagship_cache_v4 import FlagshipCacheV4Part
from cofactr.schema.flagship_cache_v5 import FlagshipCacheV5Part
from cofactr.schema.logistics import LogisticsOffer, LogisticsPart
from cofactr.schema.logistics_v2 import (
    LogisticsV2Part,
    LogisticsV2Offer,
    LogisticsV2Seller,
)
from cofactr.schema.logistics_v3 import LogisticsV3Part
from cofactr.schema.logistics_v4 import LogisticsV4Part
from cofactr.schema.price_solver_v0 import PriceSolverV0Part
from cofactr.schema.price_solver_v1 import PriceSolverV1Part
from cofactr.schema.price_solver_v2 import PriceSolverV2Part
from cofactr.schema.price_solver_v3 import PriceSolverV3Part
from cofactr.schema.price_solver_v4 import PriceSolverV4Part
from cofactr.schema.price_solver_v5 import PriceSolverV5Part
from cofactr.schema.price_solver_v6 import PriceSolverV6Part
from cofactr.schema.price_solver_v7 import PriceSolverV7Part
from cofactr.schema.price_solver_v8 import PriceSolverV8Part
from cofactr.schema.price_solver_v9 import PriceSolverV9Part


class ProductSchemaName(str, Enum):
    """Product schema name."""

    INTERNAL = "internal"
    FLAGSHIP = "flagship"
    FLAGSHIP_V2 = "flagship-v2"
    FLAGSHIP_V3 = "flagship-v3"
    FLAGSHIP_V4 = "flagship-v4"
    FLAGSHIP_V5 = "flagship-v5"
    FLAGSHIP_V6 = "flagship-v6"
    FLAGSHIP_V7 = "flagship-v7"
    FLAGSHIP_ALTS_V0 = "flagship-alts-v0"
    FLAGSHIP_CACHE_V0 = "flagship-cache-v0"
    FLAGSHIP_CACHE_V1 = "flagship-cache-v1"
    FLAGSHIP_CACHE_V2 = "flagship-cache-v2"
    FLAGSHIP_CACHE_V3 = "flagship-cache-v3"
    FLAGSHIP_CACHE_V4 = "flagship-cache-v4"
    FLAGSHIP_CACHE_V5 = "flagship-cache-v5"
    LOGISTICS = "logistics"
    LOGISTICS_V2 = "logistics-v2"
    LOGISTICS_V3 = "logistics-v3"
    LOGISTICS_V4 = "logistics-v4"
    PRICE_SOLVER_V0 = "price-solver-v0"
    PRICE_SOLVER_V1 = "price-solver-v1"
    PRICE_SOLVER_V2 = "price-solver-v2"
    PRICE_SOLVER_V3 = "price-solver-v3"
    PRICE_SOLVER_V4 = "price-solver-v4"
    PRICE_SOLVER_V5 = "price-solver-v5"
    PRICE_SOLVER_V6 = "price-solver-v6"
    PRICE_SOLVER_V7 = "price-solver-v7"
    PRICE_SOLVER_V8 = "price-solver-v8"
    PRICE_SOLVER_V9 = "price-solver-v9"


schema_to_product: Dict[ProductSchemaName, Callable] = {
    ProductSchemaName.FLAGSHIP: FlagshipPart,
    ProductSchemaName.FLAGSHIP_V2: FlagshipV2Part,
    ProductSchemaName.FLAGSHIP_V3: FlagshipV3Part,
    ProductSchemaName.FLAGSHIP_V4: FlagshipV4Part,
    ProductSchemaName.FLAGSHIP_V5: FlagshipV5Part,
    ProductSchemaName.FLAGSHIP_V6: FlagshipV6Part,
    ProductSchemaName.FLAGSHIP_V7: FlagshipV7Part,
    ProductSchemaName.FLAGSHIP_ALTS_V0: FlagshipAltsV0Part,
    ProductSchemaName.FLAGSHIP_CACHE_V0: FlagshipCacheV0Part,
    ProductSchemaName.FLAGSHIP_CACHE_V1: FlagshipCacheV1Part,
    ProductSchemaName.FLAGSHIP_CACHE_V2: FlagshipCacheV2Part,
    ProductSchemaName.FLAGSHIP_CACHE_V3: FlagshipCacheV3Part,
    ProductSchemaName.FLAGSHIP_CACHE_V4: FlagshipCacheV4Part,
    ProductSchemaName.FLAGSHIP_CACHE_V5: FlagshipCacheV5Part,
    ProductSchemaName.LOGISTICS: LogisticsPart,
    ProductSchemaName.LOGISTICS_V2: LogisticsV2Part,
    ProductSchemaName.LOGISTICS_V3: LogisticsV3Part,
    ProductSchemaName.LOGISTICS_V4: LogisticsV4Part,
    ProductSchemaName.PRICE_SOLVER_V0: PriceSolverV0Part,
    ProductSchemaName.PRICE_SOLVER_V1: PriceSolverV1Part,
    ProductSchemaName.PRICE_SOLVER_V2: PriceSolverV2Part,
    ProductSchemaName.PRICE_SOLVER_V3: PriceSolverV3Part,
    ProductSchemaName.PRICE_SOLVER_V4: PriceSolverV4Part,
    ProductSchemaName.PRICE_SOLVER_V5: PriceSolverV5Part,
    ProductSchemaName.PRICE_SOLVER_V6: PriceSolverV6Part,
    ProductSchemaName.PRICE_SOLVER_V7: PriceSolverV7Part,
    ProductSchemaName.PRICE_SOLVER_V8: PriceSolverV8Part,
    ProductSchemaName.PRICE_SOLVER_V9: PriceSolverV9Part,
}


class OfferSchemaName(str, Enum):
    """Offer schema name."""

    INTERNAL = "internal"
    FLAGSHIP = "flagship"
    FLAGSHIP_V2 = "flagship-v2"
    FLAGSHIP_V3 = "flagship-v3"
    FLAGSHIP_V4 = "flagship-v4"
    FLAGSHIP_V5 = "flagship-v5"
    FLAGSHIP_V6 = "flagship-v6"
    FLAGSHIP_V7 = "flagship-v7"
    LOGISTICS = "logistics"
    LOGISTICS_V2 = "logistics-v2"


schema_to_offer: Dict[OfferSchemaName, Callable] = {
    OfferSchemaName.INTERNAL: identity,
    OfferSchemaName.FLAGSHIP: FlagshipOffer,
    OfferSchemaName.FLAGSHIP_V2: FlagshipV2Offer,
    OfferSchemaName.FLAGSHIP_V3: FlagshipV3Offer,
    OfferSchemaName.FLAGSHIP_V4: FlagshipV4Offer,
    OfferSchemaName.FLAGSHIP_V5: FlagshipV5Offer,
    OfferSchemaName.FLAGSHIP_V6: FlagshipV6Offer,
    OfferSchemaName.FLAGSHIP_V7: FlagshipV7Offer,
    OfferSchemaName.LOGISTICS: LogisticsOffer,
    OfferSchemaName.LOGISTICS_V2: LogisticsV2Offer,
}


class OrgSchemaName(str, Enum):
    """Organization schema name."""

    INTERNAL = "internal"
    FLAGSHIP = "flagship"
    FLAGSHIP_V2 = "flagship-v2"
    FLAGSHIP_V3 = "flagship-v3"
    FLAGSHIP_V4 = "flagship-v4"
    FLAGSHIP_V5 = "flagship-v5"
    FLAGSHIP_V6 = "flagship-v6"
    LOGISTICS = "logistics"
    LOGISTICS_V2 = "logistics-v2"


schema_to_org: Dict[OrgSchemaName, Callable] = {
    OrgSchemaName.INTERNAL: identity,
    OrgSchemaName.FLAGSHIP: FlagshipSeller,
    OrgSchemaName.FLAGSHIP_V2: FlagshipV2Seller,
    OrgSchemaName.FLAGSHIP_V3: FlagshipV3Seller,
    OrgSchemaName.FLAGSHIP_V4: FlagshipV4Seller,
    OrgSchemaName.FLAGSHIP_V5: FlagshipV5Seller,
    OrgSchemaName.FLAGSHIP_V6: FlagshipV6Seller,
    OrgSchemaName.LOGISTICS: FlagshipSeller,
    OrgSchemaName.LOGISTICS_V2: LogisticsV2Seller,
}


class SupplierSchemaName(str, Enum):
    """Supplier schema name."""

    INTERNAL = "internal"
    FLAGSHIP = "flagship"
    FLAGSHIP_V2 = "flagship-v2"
    FLAGSHIP_V3 = "flagship-v3"
    FLAGSHIP_V4 = "flagship-v4"
    FLAGSHIP_V5 = "flagship-v5"
    FLAGSHIP_V6 = "flagship-v6"
    LOGISTICS = "logistics"
    LOGISTICS_V2 = "logistics-v2"


schema_to_supplier: Dict[SupplierSchemaName, Callable] = {
    SupplierSchemaName.INTERNAL: identity,
    SupplierSchemaName.FLAGSHIP: FlagshipSeller,
    SupplierSchemaName.FLAGSHIP_V2: FlagshipV2Seller,
    SupplierSchemaName.FLAGSHIP_V3: FlagshipV3Seller,
    SupplierSchemaName.FLAGSHIP_V4: FlagshipV4Seller,
    SupplierSchemaName.FLAGSHIP_V5: FlagshipV5Seller,
    SupplierSchemaName.FLAGSHIP_V6: FlagshipV6Seller,
    SupplierSchemaName.LOGISTICS: FlagshipSeller,
    SupplierSchemaName.LOGISTICS_V2: LogisticsV2Seller,
}


class OrderSchemaName(str, Enum):
    """Order schema name."""

    FLAGSHIP = "flagship"


schema_to_order: Dict[OrderSchemaName, Callable] = {
    OrderSchemaName.FLAGSHIP: FlagshipOrderStatus,
}



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import List, Optional

# Local Modules
from cofactr.schema.flagship_v4.part import Part as FlagshipV4Part


@dataclass
class Part(FlagshipV4Part):
    """Part."""

    owner_id: Optional[str]
    custom_id: Optional[str]
    deprecated_custom_ids: List[str]



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import Optional


@dataclass
class Part:  # pylint: disable=too-many-instance-attributes
    """Part."""

    id: str

    created_at: str
    modified_at: str

    mpn: Optional[str]
    mfr: Optional[str]
    hero_image: Optional[str]
    classification: Optional[str]
    description: Optional[str]
    package: Optional[str]
    terminations: Optional[int]
    termination_type: str
    lifecycle_status: str



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import Optional

# Local Modules
from cofactr.schema.flagship_v3.part import Part as FlagshipV3Part


@dataclass
class Part(FlagshipV3Part):
    """Part."""

    lifecycle_status: Optional[str]



"""Part seller class."""
# Standard Modules
from dataclasses import dataclass
from typing import Optional

# Local Modules
from cofactr.schema.flagship.seller import Seller as FlagshipSeller

@dataclass
class Seller(FlagshipSeller):
    """Part seller."""

    best_case_lead: Optional[int]



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import List

# Local Modules
from cofactr.schema.flagship_v4.offer import Offer as FlagshipV4Offer
from cofactr.schema.flagship_v7.part import Part as FlagshipV7Part


@dataclass
class Part(FlagshipV7Part):
    """Part."""

    offers: List[FlagshipV4Offer]

    def __post_init__(self):
        """Post initialization."""

        self.mfg = self.mfr
        self.offers = [FlagshipV4Offer(**offer) for offer in self.offers]



"""Part seller class."""
# Standard Modules
from dataclasses import dataclass

# Local Modules
from cofactr.schema.flagship_v4.seller import Seller as FlagshipV4Seller


@dataclass
class Seller(FlagshipV4Seller):
    """Part seller."""

    # Does the API support placing new orders for this seller?
    is_api_ordering_supported: bool
    # Does the API support checking the status of existing orders for this seller?
    is_api_order_status_supported: bool



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import List

# Local Modules
from cofactr.schema.flagship_v2.offer import Offer as FlagshipV2Offer
from cofactr.schema.flagship_v4.part import Part as FlagshipV4Part


@dataclass
class Part(FlagshipV4Part):
    """Part."""

    offers: List[FlagshipV2Offer]

    def __post_init__(self):
        """Post initialization."""

        self.mfg = self.mfr
        self.offers = [FlagshipV2Offer(**offer) for offer in self.offers]



"""Types for use in schema definitions."""
# Standard Modules
from typing import Any, List, Literal, Optional, TypedDict
from typing_extensions import NotRequired


class OfferCorrection(TypedDict):
    """Offer correction."""

    correct_value: Any
    original_value: Any
    detail: Optional[str]
    created_at: str


class Document(TypedDict):
    """Document."""

    label: str
    url: str
    filename: str


class Completion(TypedDict):
    """Autocomplete prediction."""

    id: str
    label: str


LifecycleStatusType = Literal[
    "New",
    "NRFND",
    "Production",
    "EOL",
    "Obsolete",
]
TerminationType = Literal[
    "other",
    "SMT",
    "THT",
    "pressed fit",
    "hybrid of SMT and THT",
    "hybrid of pressed fit and SMT",
    "hybrid of pressed fit and THT",
]


class ManufacturerInV0(TypedDict):
    """Manufacturer input."""

    custom_label: NotRequired[Optional[str]]
    custom_id: NotRequired[Optional[str]]
    id: NotRequired[Optional[str]]


class ClassificationInV0(TypedDict):
    """Classification input."""

    custom_label: NotRequired[Optional[str]]
    custom_id: NotRequired[Optional[str]]
    id: NotRequired[Optional[str]]


class PartInV0(TypedDict):
    """Part input."""

    owner_id: str

    mpn: str
    alt_mpns: NotRequired[Optional[List[str]]]
    custom_id: NotRequired[Optional[str]]
    mfr: NotRequired[Optional[ManufacturerInV0]]

    classification: NotRequired[Optional[ClassificationInV0]]
    description: NotRequired[Optional[str]]
    msl: NotRequired[Optional[str]]
    package: NotRequired[Optional[str]]
    terminations: NotRequired[Optional[int]]
    termination_type: NotRequired[Optional[TerminationType]]


class PartialPartInV0(TypedDict):
    """Partial part input."""

    owner_id: NotRequired[Optional[str]]

    mpn: NotRequired[Optional[List[str]]]
    alt_mpns: NotRequired[Optional[List[str]]]
    custom_id: NotRequired[Optional[str]]
    mfr: NotRequired[Optional[ManufacturerInV0]]

    classification: NotRequired[Optional[ClassificationInV0]]
    description: NotRequired[Optional[str]]
    lifecycle_status: NotRequired[Optional[LifecycleStatusType]]
    msl: NotRequired[Optional[str]]
    package: NotRequired[Optional[str]]
    terminations: NotRequired[Optional[int]]
    termination_type: NotRequired[Optional[TerminationType]]


class ScheduledReleaseV0(TypedDict):
    """Scheduled release."""

    scheduled_date: Optional[str]  # Datetime.
    scheduled_quantity: Optional[int]


class OrderLineInV0(TypedDict):
    """Order line input."""

    customer_reference: str
    cofactr_product_id: str
    seller_product_id: str
    quantity_ordered: int
    # Expected unit price. This is required so any discrepancies between between the expected and
    # actual price can be addressed.
    expected_unit_price: float
    schedule: List[ScheduledReleaseV0]


class PostalAddressV0(TypedDict):
    """Postal Address."""

    # The country (Ex: `"USA"`). You can also provide the two-letter ISO 3166-1 alpha-2 country
    # code (Ex: `"US"`).
    country: str
    # The region in which the locality is, and which is in the country (Ex: `"California"` or
    # another appropriate first-level Administrative division).
    region: str
    # The locality in which the street address is, and which is in the region
    # (Ex: `"Mountain View"`).
    locality: str
    # The postal code (Ex: `"94043"`).
    postal_code: str
    # Attention line (Ex: `"Engineering Department"`).
    attention_line: NotRequired[Optional[str]]
    # Street address line one (Ex: `"1000 Main Street"`).
    street_address_line_one: str
    # Street address line two (Ex: `"Suite 300"`).
    street_address_line_two: NotRequired[Optional[str]]


class ContactV0(TypedDict):
    """Contact."""

    # Customer ID in seller's system.
    seller_customer_id: str
    company: str
    first_name: str
    last_name: str
    telephone: str
    email: str
    address: PostalAddressV0


class OrderInV0(TypedDict):
    """Order input."""

    # Order owner, which dictates what configuration will be used to place the order. If it is
    # `None` or not provided, Cofactr's own distributor accounts and configuration will be used.
    # If an ID is provided that doesn't have configuration set up, an error will be returned.
    owner_id: NotRequired[Optional[str]]

    # Seller ID.
    seller_id: str
    # Purchase order number.
    po_number: str
    buyer_contact: ContactV0
    shipping_contact: ContactV0
    # Billing account ID (Example: Digi-Key Net Terms Billing account number). If it is `None`
    # or not provided, the fall back will be the default set up in integration-wide configuration.
    billing_account_id: NotRequired[Optional[str]]
    shipment_grouping_preference: Literal[
        "unspecified", "as-available", "min-shipments"
    ]
    # Shipping methods in order of preference (Example: `["FedEx Ground"]`).
    shipping_methods: List[str]
    order_lines: List[OrderLineInV0]



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import Optional

# Local Modules
from cofactr.schema.flagship_v6.part import Part as FlagshipV6Part


@dataclass
class Part(FlagshipV6Part):
    """Part."""

    fresh_data_at: Optional[str]



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import List, Literal, Optional, TypedDict

Source = Literal["nexar", "digikey"]

class SuggestedAlt(TypedDict):
    """Offer correction."""

    id: str
    sources: List[Source]

@dataclass
class Part:  # pylint: disable=too-many-instance-attributes
    """Part."""

    owner_id: Optional[str]
    id: str
    deprecated_ids: List[str]
    suggested_alts: List[SuggestedAlt]



"""Part offer class."""
# Standard Modules
from dataclasses import dataclass
from typing import Optional

# Local Modules
from cofactr.schema.flagship_v4.offer import Offer as FlagshipV4Offer
from cofactr.schema.flagship_v4.seller import Seller as FlagshipV4Seller


@dataclass
class Offer(FlagshipV4Offer):
    """Part offer."""

    seller: FlagshipV4Seller
    convenience_return_window: Optional[int]  # Days.
    scheduled_release_period: Optional[int]  # Months.

    def __post_init__(self):
        """Convert types."""

        self.seller = FlagshipV4Seller(**self.seller)  # pylint: disable=not-a-mapping



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import List, Optional

# Local Modules
from cofactr.schema.flagship_cache_v1.part import Part as FlagshipCacheV1Part


@dataclass
class Part(FlagshipCacheV1Part):  # pylint: disable=too-many-instance-attributes
    """Part."""

    owner_id: Optional[str]
    custom_id: Optional[str]
    deprecated_custom_ids: List[str]



"""Cofactr graph API client."""
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# Python Modules
from enum import Enum
import json
from typing import Any, Dict, List, Literal, NamedTuple, Optional
from urllib.parse import quote, urlencode

# 3rd Party Modules
import httpx
from more_itertools import batched, flatten
from tenacity import (
    retry,
    retry_any,
    retry_if_exception_message,
    retry_if_exception_type,
    stop_after_attempt,
    wait_chain,
    wait_fixed,
)

# Local Modules
from cofactr.helpers import parse_entities
from cofactr.schema import (
    OfferSchemaName,
    OrderSchemaName,
    OrgSchemaName,
    ProductSchemaName,
    SupplierSchemaName,
    schema_to_offer,
    schema_to_order,
    schema_to_org,
    schema_to_product,
    schema_to_supplier,
)
from cofactr.schema.types import Completion, OrderInV0, PartInV0, PartialPartInV0

Protocol = Literal["http", "https"]

_MAX_BATCH_SIZE = 250
_MAX_SUB_BATCH_SIZE = 25


class SearchStrategy(str, Enum):
    """Search strategy."""

    DEFAULT = "default"
    MPN_SKU_MFR = "mpn_sku_mfr"


drop_none_values = lambda d: {k: v for k, v in d.items() if v is not None}

BATCH_LIMIT = 500


class RetrySettings(NamedTuple):
    """Retry settings for GraphAPI methods.
    TODO: Consider extending to other 5xx errors if and when encountered.
    """

    reraise: bool = True
    retry: retry_any = (
        retry_if_exception_type(httpx.ConnectTimeout)
        | retry_if_exception_type(httpx.ReadTimeout)
        | retry_if_exception_message(match=r"Server error '502 Bad Gateway'")
    )
    stop: stop_after_attempt = stop_after_attempt(3)
    wait: wait_chain = wait_chain(*[wait_fixed(wait=wait) for wait in [1, 3, 5]])


def get_products(
    url,
    client_id,
    api_key,
    query,
    fields,
    before,
    after,
    limit,
    external,
    force_refresh,
    schema,
    filtering,
    search_strategy: SearchStrategy,
    stale_delta: Optional[str],
    timeout: Optional[int] = None,
    owner_id: Optional[str] = None,
    reference: Optional[str] = None,
) -> httpx.Response:
    """Get products."""

    res = httpx.get(
        f"{url}/products/",
        headers=drop_none_values(
            {
                "X-CLIENT-ID": client_id,
                "X-API-KEY": api_key,
            }
        ),
        params=drop_none_values(
            {
                "owner_id": owner_id,
                "q": query,
                "fields": fields,
                "before": before,
                "after": after,
                "limit": limit,
                "external": external,
                "force_refresh": force_refresh,
                "schema": schema,
                "filtering": json.dumps(filtering) if filtering else None,
                "search_strategy": search_strategy.value,
                "stale_delta": stale_delta,
                "ref": reference,
            }
        ),
        timeout=timeout,
        follow_redirects=True,
    )

    res.raise_for_status()

    return res


def get_orgs(
    url,
    client_id,
    api_key,
    query,
    before,
    after,
    limit,
    schema,
    timeout,
    filtering,
    owner_id,
) -> httpx.Response:
    """Get orgs."""

    res = httpx.get(
        f"{url}/orgs",
        headers=drop_none_values(
            {
                "X-CLIENT-ID": client_id,
                "X-API-KEY": api_key,
            }
        ),
        params=drop_none_values(
            {
                "owner_id": owner_id,
                "q": query,
                "before": before,
                "after": after,
                "limit": limit,
                "schema": schema,
                "filtering": json.dumps(filtering) if filtering else None,
            }
        ),
        timeout=timeout,
        follow_redirects=True,
    )

    res.raise_for_status()

    return res


def get_suppliers(
    url,
    client_id,
    api_key,
    query,
    before,
    after,
    limit,
    schema,
    timeout,
    filtering,
    owner_id,
) -> httpx.Response:
    """Get orgs."""

    res = httpx.get(
        f"{url}/orgs/suppliers",
        headers=drop_none_values(
            {
                "X-CLIENT-ID": client_id,
                "X-API-KEY": api_key,
            }
        ),
        params=drop_none_values(
            {
                "owner_id": owner_id,
                "q": query,
                "before": before,
                "after": after,
                "limit": limit,
                "schema": schema,
                "filtering": json.dumps(filtering) if filtering else None,
            }
        ),
        timeout=timeout,
        follow_redirects=True,
    )

    res.raise_for_status()

    return res


class GraphAPI:  # pylint: disable=too-many-instance-attributes
    """A client-side representation of the Cofactr graph API."""

    PROTOCOL: Protocol = "https"
    HOST = "graph.cofactr.com"
    retry_settings = RetrySettings()

    def __init__(
        self,
        protocol: Optional[Protocol] = PROTOCOL,
        host: Optional[str] = HOST,
        default_product_schema: ProductSchemaName = ProductSchemaName.FLAGSHIP,
        default_order_schema: OrderSchemaName = OrderSchemaName.FLAGSHIP,
        default_org_schema: OrgSchemaName = OrgSchemaName.FLAGSHIP,
        default_offer_schema: OfferSchemaName = OfferSchemaName.FLAGSHIP,
        default_supplier_schema: SupplierSchemaName = SupplierSchemaName.FLAGSHIP,
        client_id: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        self.url = f"{protocol}://{host}"
        self.default_product_schema = default_product_schema
        self.default_order_schema = default_order_schema
        self.default_org_schema = default_org_schema
        self.default_offer_schema = default_offer_schema
        self.default_supplier_schema = default_supplier_schema
        self.client_id = client_id
        self.api_key = api_key

    def check_health(self):
        """Check the operational status of the service."""

        res = httpx.get(self.url)

        res.raise_for_status()

        return res.json()

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_products(
        self,
        query: Optional[str] = None,
        fields: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        external: Optional[bool] = True,
        force_refresh: bool = False,
        schema: Optional[ProductSchemaName] = None,
        filtering: Optional[List[Dict]] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        search_strategy: SearchStrategy = SearchStrategy.DEFAULT,
        stale_delta: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Get products.

        Args:
            query: Search query.
            fields: Used to filter properties that the response should contain. A field can be a
                concrete property like "mpn" or an abstract group of properties like "assembly".
                Example: `"id,aliases,labels,statements{spec,assembly},offers"`.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            external: Whether to query external sources.
            force_refresh: Whether to force re-ingestion from external sources. Overrides
                `external`.
            schema: Response schema.
            filtering: Filter products.
                Example: `[{"field":"id","operator":"IN","value":["CCCQSA3G9SMR","CCV1F7A8UIYH"]}]`.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            search_strategy: Search strategy used to find products.
            stale_delta: How much time has to pass before data is treated as stale. Use "inf" or
                "infinite" to indicate data should not be refreshed, no matter how old.
                Examples: "5h", "1d", "1w", "inf"
            reference: Arbitrary note to associate with the request.
        """

        if not schema:
            schema = self.default_product_schema

        res = get_products(
            url=self.url,
            client_id=self.client_id,
            api_key=self.api_key,
            query=query,
            fields=fields,
            external=external,
            force_refresh=force_refresh,
            before=before,
            after=after,
            limit=limit,
            schema=schema.value,
            filtering=filtering,
            search_strategy=search_strategy,
            stale_delta=stale_delta,
            timeout=timeout,
            owner_id=owner_id,
            reference=reference,
        )

        extracted_products = res.json()

        # Handle schemas that have parsers.
        Product = schema_to_product.get(schema)  # pylint: disable=invalid-name

        if Product:
            extracted_products["data"] = [
                Product(**data) for data in extracted_products["data"]
            ]

        return extracted_products

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_products_by_searches(
        self,
        queries: List[str],
        external: bool = True,
        force_refresh: bool = False,
        schema: Optional[ProductSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        search_strategy: SearchStrategy = SearchStrategy.DEFAULT,
        stale_delta: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Search for products associated with each query.

        Args:
            queries: Queries to find products for.
            external: Whether to query external sources in order to refresh data if applicable.
            force_refresh: Whether to force re-ingestion from external sources. Overrides
                `external`.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            search_strategy: Search strategy used to find products.
            stale_delta: How much time has to pass before data is treated as stale. Use "inf" or
                "infinite" to indicate data should not be refreshed, no matter how old.
                Examples: "5h", "1d", "1w", "inf"
            reference: Arbitrary note to associate with the request.

        Returns:
            A dictionary mapping each MPN to a list of matching products.
        """

        if not queries:
            return {}

        if not schema:
            schema = self.default_product_schema

        stale_delta_param = f"&stale_delta={stale_delta}" if stale_delta else ""

        query_to_products: Dict[str, Any] = {}

        for query_batch in batched(queries, n=_MAX_BATCH_SIZE):
            res = httpx.post(
                f"{self.url}/batch/products/",
                headers=drop_none_values(
                    {
                        "X-CLIENT-ID": self.client_id,
                        "X-API-KEY": self.api_key,
                    }
                ),
                json={
                    "batch": [
                        {
                            "method": "GET",
                            "relative_url": (
                                f"?q={quote(query)}&schema={schema.value}&external={bool(external)}"
                                f"&force_refresh={force_refresh}{stale_delta_param}"
                                f"&search_strategy={search_strategy.value}"
                            ),
                        }
                        for query in query_batch
                    ]
                },
                params=drop_none_values({"owner_id": owner_id, "ref": reference}),
                timeout=timeout,
                follow_redirects=True,
            )

            res.raise_for_status()

            responses = res.json()

            Product = schema_to_product[schema]  # pylint: disable=invalid-name

            for query, response in zip(query_batch, responses):
                matches = []

                if response["code"] == 200:
                    data = response["body"]["data"]

                    matches = [Product(**product_data) for product_data in data]

                query_to_products[query] = matches

        return query_to_products

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_products_by_ids(
        self,
        ids: List[str],
        external: Optional[bool] = True,
        force_refresh: bool = False,
        schema: Optional[ProductSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        stale_delta: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Get a batch of products by IDs.

        Note: Multiple requests are made if more than 250 IDs are provided.

        Args:
            ids: Cofactr product IDs to match on.
            external: Whether to query external sources in order to refresh data if applicable.
            force_refresh: Whether to force re-ingestion from external sources. Overrides
                `external`.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            stale_delta: How much time has to pass before data is treated as stale. Use "inf" or
                "infinite" to indicate data should not be refreshed, no matter how old.
                Examples: "5h", "1d", "1w", "inf"
            reference: Arbitrary note to associate with the request.
        """

        if not ids:
            return {}

        if not schema:
            schema = self.default_product_schema

        batched_products = [
            self.get_products(
                external=external,
                force_refresh=force_refresh,
                schema=schema,
                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
                limit=_MAX_BATCH_SIZE,
                timeout=timeout,
                owner_id=owner_id,
                stale_delta=stale_delta,
                reference=reference,
            )
            for batched_ids in batched(ids, n=_MAX_BATCH_SIZE)
        ]

        products_data = list(flatten([res["data"] for res in batched_products]))

        id_to_product = parse_entities(
            ids=ids,
            entities=products_data,
            entity_dataclass=schema_to_product[schema],
        )

        return id_to_product

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_canonical_product_ids(
        self,
        ids: List[str],
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Get the canonical product ID for each of the given IDs, which may or may not be
        deprecated.
        """

        if not ids:
            return {}

        batched_products = [
            self.get_products(
                fields="id,deprecated_ids",
                external=False,
                force_refresh=False,
                schema=ProductSchemaName.INTERNAL,
                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
                limit=_MAX_BATCH_SIZE,
                timeout=timeout,
                owner_id=owner_id,
                reference=reference,
            )
            for batched_ids in batched(ids, n=_MAX_BATCH_SIZE)
        ]

        id_to_canonical_id = {}

        for res in batched_products:
            products = res.get("data", [])

            for product in products:
                canonical_id = product["id"]

                for key in [canonical_id, *(product.get("deprecated_ids") or [])]:
                    id_to_canonical_id[key] = canonical_id

        return {id_: id_to_canonical_id.get(id_) for id_ in ids}

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_orgs(
        self,
        query: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        schema: Optional[OrgSchemaName] = None,
        timeout: Optional[int] = None,
        filtering: Optional[List[Dict]] = None,
        owner_id: Optional[str] = None,
    ):
        """Get organizations.

        Args:
            query: Search query.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            filtering: Filter orgs.
                Example: `[{"field":"id","operator":"IN","value":["622fb450e4c292d8287b0af5"]}]`.
            owner_id: Specifies which private data to access.
        """

        if not schema:
            schema = self.default_org_schema

        res = get_orgs(
            url=self.url,
            client_id=self.client_id,
            api_key=self.api_key,
            query=query,
            before=before,
            after=after,
            limit=limit,
            schema=schema.value,
            timeout=timeout,
            filtering=filtering,
            owner_id=owner_id,
        )

        res_json = res.json()

        Org = schema_to_org[schema]  # pylint: disable=invalid-name

        res_json["data"] = [Org(**data) for data in res_json["data"]]

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_suppliers(
        self,
        query: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = None,
        schema: Optional[SupplierSchemaName] = None,
        timeout: Optional[int] = None,
        filtering: Optional[List[Dict]] = None,
        owner_id: Optional[str] = None,
    ):
        """Get suppliers.

        Args:
            query: Search query.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of documents.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            filtering: Filter suppliers.
                Example: `[{"field":"id","operator":"IN","value":["622fb450e4c292d8287b0af5"]}]`.
            owner_id: Specifies which private data to access.
        """

        if not schema:
            schema = self.default_supplier_schema

        res = get_suppliers(
            url=self.url,
            client_id=self.client_id,
            api_key=self.api_key,
            query=query,
            before=before,
            after=after,
            limit=limit,
            schema=schema.value,
            timeout=timeout,
            filtering=filtering,
            owner_id=owner_id,
        )

        res_json = res.json()

        Supplier = schema_to_supplier[schema]  # pylint: disable=invalid-name

        res_json["data"] = [Supplier(**data) for data in res_json["data"]]

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_suppliers_by_ids(
        self,
        ids: List[str],
        schema: Optional[SupplierSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
    ):
        """Get a batch of suppliers by IDs.

        Note: Multiple requests are made if more than 250 IDs are provided.

        Args:
            ids: Cofactr org IDs to match on.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
        """

        if not ids:
            return {}

        if not schema:
            schema = self.default_supplier_schema

        batched_suppliers = [
            self.get_suppliers(
                schema=schema,
                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
                limit=_MAX_BATCH_SIZE,
                timeout=timeout,
                owner_id=owner_id,
            )
            for batched_ids in batched(ids, n=_MAX_BATCH_SIZE)
        ]

        suppliers_data = list(
            flatten([suppliers["data"] for suppliers in batched_suppliers])
        )

        id_to_supplier = parse_entities(
            ids=ids,
            entities=suppliers_data,
            entity_dataclass=schema_to_supplier[schema],
        )

        return id_to_supplier

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def autocomplete_orgs(
        self,
        query: Optional[str] = None,
        limit: Optional[int] = None,
        types: Optional[str] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
    ) -> Dict[Literal["data"], Completion]:
        """Autocomplete organizations.

        Args:
            query: Search query.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of
                documents.
            types: Filter for types of organizations.
                Example: "supplier" filters to suppliers.
                Example: "supplier|manufacturer" filters to orgs that are a
                    supplier or a manufacturer.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
        """

        res = httpx.get(
            f"{self.url}/orgs/autocompletions/",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values(
                {
                    "owner_id": owner_id,
                    "q": query,
                    "limit": limit,
                    "types": types,
                }
            ),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        return res.json()

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def autocomplete_classifications(
        self,
        query: Optional[str] = None,
        limit: Optional[int] = None,
        types: Optional[str] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
    ) -> Dict[Literal["data"], Completion]:
        """Autocomplete classifications.

        Args:
            query: Search query.
            before: Upper page boundry, expressed as a product ID.
            after: Lower page boundry, expressed as a product ID.
            limit: Restrict the results of the query to a particular number of
                documents.
            types: Filter for types of organizations.
                Example: "part_classification" filters to part classification classes.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
        """

        res = httpx.get(
            f"{self.url}/classes/autocompletions/",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values(
                {
                    "owner_id": owner_id,
                    "q": query,
                    "limit": limit,
                    "types": types,
                }
            ),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        return res.json()

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_product(
        self,
        id: str,
        fields: Optional[str] = None,
        external: Optional[bool] = True,
        force_refresh: bool = False,
        schema: Optional[ProductSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        stale_delta: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Get product.

        Args:
            fields: Used to filter properties that the response should contain. A field can be a
                concrete property like "mpn" or an abstract group of properties like "assembly".
                Example: "id,aliases,labels,statements{spec,assembly},offers"
            external: Whether to query external sources in order to update information for the
                given product.
            force_refresh: Whether to force re-ingestion from external sources. Overrides
                `external`.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            stale_delta: How much time has to pass before data is treated as stale. Use "inf" or
                "infinite" to indicate data should not be refreshed, no matter how old.
                Examples: "5h", "1d", "1w", "inf"
            reference: Arbitrary note to associate with the request.
        """

        if not schema:
            schema = self.default_product_schema

        res = httpx.get(
            f"{self.url}/products/{id}",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values(
                {
                    "owner_id": owner_id,
                    "fields": fields,
                    "external": external,
                    "force_refresh": force_refresh,
                    "schema": schema.value,
                    "stale_delta": stale_delta,
                    "ref": reference,
                }
            ),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        res_json = res.json()

        Product = schema_to_product[schema]  # pylint: disable=invalid-name

        res_json["data"] = (
            Product(**res_json["data"]) if (res_json and res_json.get("data")) else None
        )

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def create_product(
        self,
        data: PartInV0,
        schema: Optional[ProductSchemaName] = None,
        timeout: Optional[int] = None,
        reference: Optional[str] = None,
    ):
        """Create product.

        Args:
            data: Data defining the product to create.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            reference: Arbitrary note to associate with the request.

        Returns:
            The newly created product, represented in the given schema.
        """

        if not schema:
            schema = self.default_product_schema

        res = httpx.post(
            f"{self.url}/products/",
            json=data,
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values({"schema": schema.value, "ref": reference}),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        res_json = res.json()

        Product = schema_to_product[schema]  # pylint: disable=invalid-name

        res_json["data"] = (
            Product(**res_json["data"]) if (res_json and res_json.get("data")) else None
        )

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def update_product(
        self,
        product_id: str,
        data: PartialPartInV0,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Update product.

        Args:
            product_id: Cofactr ID of product to update.
            data: Data defining product updates. A value of `None` represents delete.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Data owner ID.
            reference: Arbitrary note to associate with the request.
        """

        res = httpx.patch(
            f"{self.url}/products/{product_id}",
            json={
                "owner_id": owner_id,
                "schema": "flagship",
                "data": data,
            },
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values({"ref": reference}),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def set_custom_product_ids(
        self,
        id_to_custom_id: Dict[str, Optional[str]],
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Set custom product IDs.

        Args:
            id_to_custom_id: Map from Cofactr product ID to desired custom ID.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Data owner ID.
            reference: Arbitrary note to associate with the request.
        """

        if not id_to_custom_id:
            return

        res = httpx.post(
            f"{self.url}/actions/custom-product-id-mappings/",
            json={
                "owner_id": owner_id,
                "id_to_custom_id": id_to_custom_id,
            },
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values({"ref": reference}),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_offers(
        self,
        product_id: str,
        fields: Optional[str] = None,
        external: Optional[bool] = True,
        force_refresh: bool = False,
        schema: Optional[OfferSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        stale_delta: Optional[str] = None,
        reference: Optional[str] = None,
    ):
        """Get product.

        Args:
            product_id: ID of the product to get offers for.
            fields: Used to filter properties that the response should contain.
            external: Whether to query external sources in order to update information.
            force_refresh: Whether to force re-ingestion from external sources. Overrides
                `external`.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            stale_delta: How much time has to pass before data is treated as stale. Use "inf" or
                "infinite" to indicate data should not be refreshed, no matter how old.
                Examples: "5h", "1d", "1w", "inf"
            reference: Arbitrary note to associate with the request.
        """

        if not schema:
            schema = self.default_offer_schema

        res = httpx.get(
            f"{self.url}/products/{product_id}/offers",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values(
                {
                    "owner_id": owner_id,
                    "fields": fields,
                    "external": external,
                    "force_refresh": force_refresh,
                    "schema": schema.value,
                    "stale_delta": stale_delta,
                    "ref": reference,
                }
            ),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        res_json = res.json()

        Offer = schema_to_offer[schema]  # pylint: disable=invalid-name

        res_json["data"] = [Offer(**data) for data in res_json["data"]]

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_org(
        self,
        id: str,
        schema: Optional[OrgSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
    ):
        """Get organization."""

        if not schema:
            schema = self.default_org_schema

        res = httpx.get(
            f"{self.url}/orgs/{id}",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values({"owner_id": owner_id, "schema": schema.value}),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        res_json = res.json()

        Org = schema_to_org[schema]  # pylint: disable=invalid-name

        res_json["data"] = (
            Org(**res_json["data"]) if (res_json and res_json.get("data")) else None
        )

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_supplier(
        self,
        id: str,
        schema: Optional[SupplierSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
    ):
        """Get supplier."""

        if not schema:
            schema = self.default_supplier_schema

        res = httpx.get(
            f"{self.url}/orgs/{id}",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values({"owner_id": owner_id, "schema": schema.value}),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        res_json = res.json()

        Supplier = schema_to_supplier[schema]  # pylint: disable=invalid-name

        res_json["data"] = (
            Supplier(**res_json["data"])
            if (res_json and res_json.get("data"))
            else None
        )

        return res_json

    @retry(
        reraise=retry_settings.reraise,
        retry=retry_settings.retry,
        stop=retry_settings.stop,
        wait=retry_settings.wait,
    )
    def get_orders(
        self,
        schema: Optional[OrderSchemaName] = None,
        timeout: Optional[int] = None,
        filtering: Optional[List[Dict]] = None,
        owner_id: Optional[str] = None,
        is_sandbox: bool = False,
    ):
        """Get orders.

        Args:
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            filtering: Filter orders.
                Example: `[{"field":"id","operator":"IN","value":["622fb450e4c292d8287b0af5:12345678"]}]`.
            owner_id: Specifies which private data to access.
            is_sandbox: If True, the order will be executed in a sandbox environment: Real orders
                will not be queried.
        """

        if not schema:
            schema = self.default_order_schema

        res = httpx.get(
            f"{self.url}/orders/",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            params=drop_none_values(
                {
                    "owner_id": owner_id,
                    "external": True,
                    "schema": schema,
                    "filtering": json.dumps(filtering) if filtering else None,
                    "is_sandbox": is_sandbox,
                }
            ),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        res_json = res.json()

        Order = schema_to_order[schema]  # pylint: disable=invalid-name

        res_json["data"] = [Order(**data) for data in res_json["data"]]

        return res_json

    def get_orders_by_ids(
        self,
        ids: List[str],
        schema: Optional[OrderSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        is_sandbox: bool = False,
    ):
        """Get a batch of orders by IDs.

        Args:
            ids: Cofactr order IDs to match on.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            is_sandbox: If True, the order will be executed in a sandbox environment: Real orders
                will not be queried.
        """

        if not ids:
            return {}

        if not schema:
            schema = self.default_order_schema

        batched_orders = [
            self.get_orders(
                schema=schema,
                filtering=[{"field": "id", "operator": "IN", "value": batched_ids}],
                timeout=timeout,
                owner_id=owner_id,
                is_sandbox=is_sandbox,
            )
            for batched_ids in batched(ids, n=_MAX_BATCH_SIZE)
        ]

        orders_data = list(flatten([res["data"] for res in batched_orders]))

        # All order schemas have `id` field.
        return {order.id: order for order in orders_data}

    def create_order(
        self,
        data: OrderInV0,
        timeout: Optional[int] = None,
        is_sandbox: bool = False,
    ):
        """Create order.

        Args:
            data: Data defining the order to create.
            timeout: Time to wait (in seconds) for the server to issue a response.
            is_sandbox: If True, the order will be executed in a sandbox environment: A real order
                will not be placed.

        Returns:
            ID of the created order.
        """

        res = httpx.post(
            f"{self.url}/orders/",
            headers=drop_none_values(
                {
                    "X-CLIENT-ID": self.client_id,
                    "X-API-KEY": self.api_key,
                }
            ),
            json=data,
            params=drop_none_values(
                {
                    "is_sandbox": is_sandbox,
                }
            ),
            timeout=timeout,
            follow_redirects=True,
        )

        res.raise_for_status()

        location = res.headers.get("location")

        if not location:
            raise ValueError("No resource location found in order creation response.")

        # Remove `/orders/` to get the ID from the resource path.
        return location[8:]

    def create_get_products_by_ids_job(
        self,
        ids: List[str],
        external: Optional[bool] = True,
        force_refresh: bool = False,
        schema: Optional[ProductSchemaName] = None,
        timeout: Optional[int] = None,
        owner_id: Optional[str] = None,
        stale_delta: Optional[str] = None,
        reference: Optional[str] = None,
    ) -> List[str]:
        """Create batch product request job.

        Note: Multiple jobs are created if more than 250 IDs are provided.

        Args:
            ids: Cofactr product IDs to match on.
            external: Whether to query external sources in order to refresh data if applicable.
            force_refresh: Whether to force re-ingestion from external sources. Overrides
                `external`.
            schema: Response schema.
            timeout: Time to wait (in seconds) for the server to issue a response.
            owner_id: Specifies which private data to access.
            stale_delta: How much time has to pass before data is treated as stale. Use "inf" or
                "infinite" to indicate data should not be refreshed, no matter how old.
                Examples: "5h", "1d", "1w", "inf"
            reference: Arbitrary note to associate with the request.

        Returns:
            A list with one ID for each job that was created.
        """

        if not schema:
            schema = self.default_product_schema

        invariant_query_params = drop_none_values(
            {
                "schema": schema.value,
                "external": bool(external),
                "force_refresh": force_refresh,
                "stale_delta": stale_delta,
            }
        )

        job_ids = []

        for id_batch in batched(ids, n=_MAX_BATCH_SIZE):
            res = httpx.post(
                f"{self.url}/jobs/batch-products-requests/",
                headers=drop_none_values(
                    {
                        "X-CLIENT-ID": self.client_id,
                        "X-API-KEY": self.api_key,
                    }
                ),
                json={
                    "batch": [
                        {
                            "method": "GET",
                            "relative_url": (
                                f'?filtering={filtering}&{urlencode(invariant_query_params)}'
                            ),
                        }
                        for ids_ in batched(id_batch, n=_MAX_SUB_BATCH_SIZE)
                        if (
                            filtering := quote(
                                json.dumps([{"field": "id", "operator": "IN", "value": ids_}])
                            )
                        )
                    ]
                },
                params=drop_none_values({"owner_id": owner_id, "ref": reference}),
                timeout=timeout,
                follow_redirects=True,
            )
            res.raise_for_status()

            location = res.headers.get("location")

            if not location:
                raise ValueError("No resource location found in job creation response.")

            # Remove `/jobs/` to get the ID from the resource path.
            job_ids.append(location[6:])

        return job_ids



"""Order class."""
# pylint: disable=not-a-mapping
# Standard Modules
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Error:
    """Error."""

    code: Optional[str]
    message: Optional[str]
    name: Optional[str]


@dataclass
class ScheduledRelease:
    """Scheduled release."""

    scheduled_date: Optional[str]
    scheduled_quantity: Optional[int]


@dataclass
class OrderStatusLine:
    """Order status line."""

    line_id: str
    customer_reference: Optional[str]
    cofactr_product_id: Optional[str]
    seller_product_id: Optional[str]
    quantity_ordered: Optional[int]
    quantity_shipped: Optional[int]
    quantity_backordered: Optional[int]
    country_of_origin: Optional[str]
    backorder_schedule: List[ScheduledRelease]
    schedule: List[ScheduledRelease]

    def __post_init__(self):
        """Convert types."""

        self.backorder_schedule = [
            ScheduledRelease(**schedule) for schedule in self.backorder_schedule
        ]
        self.schedule = [ScheduledRelease(**schedule) for schedule in self.schedule]


@dataclass
class PackageItem:
    """Package item."""

    # Line ID of item being shipped.
    line_id: str
    # Quantity in package.
    quantity: Optional[int]


@dataclass
class PackageDetail:
    """Package detail."""

    items: List[PackageItem]
    tracking_url: Optional[str]
    tracking_number: Optional[str]
    # Name of carrier that will deliver the package.
    carrier_name: Optional[str]
    # Method to be used for shipping the order.
    shipping_method: Optional[str]

    def __post_init__(self):
        """Convert types."""

        self.items = [PackageItem(**item) for item in self.items]


@dataclass
class OrderStatus:  # pylint: disable=too-many-instance-attributes
    """Order status."""

    # Unique Cofactr-defined identifier representing the order.
    id: str
    # Identifier representing the order in seller's order management system.
    seller_order_id: str
    # Errors associated with the order.
    errors: List[Error]
    package_details: List[PackageDetail]
    order_status_lines: List[OrderStatusLine]

    def __post_init__(self):
        """Convert types."""

        self.errors = [Error(**error) for error in self.errors]
        self.package_details = [
            PackageDetail(**detail) for detail in self.package_details
        ]
        self.order_status_lines = [
            OrderStatusLine(**line) for line in self.order_status_lines
        ]



"""Part class."""
# Standard Modules
from dataclasses import dataclass
from typing import Optional

# Local Modules
from cofactr.schema.flagship_cache_v2.part import Part as FlagshipCacheV2Part


@dataclass
class Part(FlagshipCacheV2Part):  # pylint: disable=too-many-instance-attributes
    """Part."""

    min_lead: Optional[int]
